--require "defines"
--Minimum version 15.12 required for flow statistics updates


--COMPLETED: Fix 1 input bug
--COMPLETED: reduced stages, removing 
--COMPLETED: Fix for steel (i forgot checking recipes)
--COMPLETED: existing usage check
--COMPLETED: removal handling
--COMPLETED: use new method for displaying items on the production graph (15.12+)
--COMPLETED: fix power-draw amounts
--COMPLETED: Make code more pretty pretty
--COMPLETED: Pull graphics from default game content
--COMPLETED: Logistic passive-output/active-output,
--COMPLETED: Fix possible loss of ore issue when power was not enough
--COMPLETED: Power Draw matches ore consumption
--COMPLETED: Technology
--COMPLETED: Locale
--COMPLETED: make sure scripts handle multiple input/output entity types
--COMPLETED: Icons
--COMPLETED: fix circuit wire sprite location
--COMPLETED: fix furnace sometimes skipping session on fully laid compound
--COMPLETED: Fix walking distance (cannot walk past inserter?!?
--COMPLETED: OnDestroy Handler!
--COMPLETED: Build large scale test (completed with 288 of 360 possible furnaces)
--COMPLETED: start core with idle drain
--0.0.8:
--COMPLETED: switch what event triggers the removal of the group, reduced spam when removing input/outputs
--COMPLETED: Fix - energy dropping to 0 when group removed
--COMPLETED: Fix - removing the first furnace group made on a save would break table lookup functions
--COMPLETED: Fix - added pollution amount, add's 7.2 pollution for every 200KW above the idle power drain on each furnace update
--0.0.9:
--COMPLETED: Fix - Weird numbers from a setup such as 100 beacons and 1 furnace
--0.0.10:
--COMPLETED: Fix - stone and steel production

--TODO: redo the balancing mechanism for outputting to chests
--TODO: Add audio queue
--TODO: Fix Building height- inserter hands go through the building
--TODO: fix fast replace breaks entity link

--postponed: make sure entities are aligned to the same 3x3 grid on placement
--postponed: make sure entities are near each other
--the honor system must do for now...
--postponed: animated entities
--postponed: make a bunch of graphics for the possible variations of the tiles
--postponed: variation reassignment on entity removal
--postponed: variation reassignment on entity place

local compound_furnace_debug = false
--[[
	Compound furnaces: A low cpu cost method of generating plates in high quantity. 
	A fully laid out compound furnace will eat through 9600 ore/iron plates a minute and output with a 20% productivity 
	bonus. I intend to make a steel version that mimics 72 furnaces as well.

	a fully laid out electric furnace will consist of:
		- 60 furnace units (built-in prod mk3) batch size: # of craft cycles in 60 seconds, rounded up.
		- 100 speed units (built-in speed mk3)
		-1 controller (electrical link)
		-1-5 input chests (steel, requester)
		-1-5 output chests(steel, passive provider, active provider)
	each building will be 3x3 in size and a fully laid out furnace will use ~130MW of power
	the power tap will 'pre-load' the power onto the structure to guarantee the duration of the craft cycle.

	a fully laid out steel furnace will consist of:
		-72 steel furnaces
		-1 controller (coal input)
		-1-4 input chests (steel, requester)
		-1-4 output chests(steel, passive provider, active provider)
	each building will be 2x2 in size

	electric compound furnace:
	craft rate = 1.4 + (speed units * 4.7604/batch units)
	craft seconds = batch units * craft rate * 6 or 60s
	batches = floor(craft seconds / item craft rate)
	productivity bonus = batches /(1/.2) + remainder of last set.
	update frequency: 6 or 60 seconds
	
	electrical demand, electric setup:
	equivalent panamax design (9600 input iron ore):
	60480 KW from 126 beacons
	88560 KW per smelter
	149040 KW draw total

	steel compound furnace:
	craft rate = 2
	craft seconds = batch units * craft rate * (6s or 60s)
	batches = floor(craft seconds / item craft rate)
	update frequency: 6 or 60 seconds

	TODO: add additional chest types (logistic, item specific via entity breadth or script functionality)
--]]

--[[
serialization safe unique identifiers:
- LuaEntity.unit_number - this is a unique Id for most entities: a save with 200+ hours has entitiy id's up to 37 million
]]--

--a lookup table for neighboring entities in position vector
adj_lookup =
{--x,y
[1] =  {x= 2.0,y=-1.0},
[2] =  {x= 2.0,y= 0.0},
[3] =  {x= 2.0,y= 1.0},
[4] =  {x=-2.0,y=-1.0},
[5] =  {x=-2.0,y= 0.0},
[6] =  {x=-2.0,y= 1.0},
[7] =  {y= 2.0,x=-1.0},
[8] =  {y= 2.0,x= 0.0},
[9] =  {y= 2.0,x= 1.0},
[10] = {y=-2.0,x=-1.0},
[11] = {y=-2.0,x= 0.0},
[12] = {y=-2.0,x= 1.0},
}

--[[the searched places
      ***
	 *BBB*
	 *BBB*
	 *BBB*
	  ***
--]]

--coordinate plane:
--   1-
--x 8-0+2
--   +
--   y4

--a mapped table to the variation types


local variation_map = 
{
 [0] =1,
 [12]=2,
 [9] =3,
 [3] =4,
 [6] =5,
 [4] =6,
 [8] =7,
 [1] =8,
 [2] =9,
 [10]=10,--not allowed (for variation cheapness)
 [5] =11,--not allowed (for variation cheapness)
 [14]=12,
 [13]=13,
 [11]=14,
 [7] = 15,
 [15] = 16,
}

designs = {}
-- a lookup table for entities that count as inputs for a compound furnace
 default_input = {}
--a lookup table for entities that count as outputs for a compound furnace
 default_output= {}

--a lookup table for entities that this mod handles
 design_lookup = {}

 design_lookup =
{
	--defaults (pointing to default_designs[1])
	["cf-elec-core"]=1,--electric control unit
	["cf-elec-batch"]=1,--furnace analogue "batch processor"
	["cf-elec-speed"]=1,--speed beacon analogue "speed processor"
	["cf-elec-input"]=1,--input chest (48 inventory slots)
	["cf-elec-requester"]=1,
	["cf-elec-output"]=1,--output chest (48 inventory slots)
	["cf-elec-passive-provider"]=1,

}

	--entities considered as inputs
 default_input= {
	"cf-elec-input",
	"cf-elec-requester",
}
	--entities considered as outputs
 default_output= {
	"cf-elec-output",
	"cf-elec-passive-provider",
}

--electrical furnace set
 designs[1] =
{
	category="smelting",
	core = "cf-elec-core",
	batch = "cf-elec-batch",
	speed ="cf-elec-speed",
	inputs = default_input,
	outputs = default_output,
	input_limit = 4,
	output_limit = 4,
	batch_limit = 60,
	batch_watt = 180,
	craft_base =1.4,--2x MK3 productivity built in
	productivity =5,--2x MK3 productivity built in, 5 cycles for 1 free cycle, max 1 free cycle per cycle
	speed_limit = 100,--max number of speed increasers
	speed_mod = 0.07934,--craft rate increase to base craft rate (linear)
	speed_watt = 300,
	grid = 3, -- the alignment size of the building
		name = "Electric Compound Furnace"
		
	--40x300 + 40x180 
}

function on_mod_changes(data)
	buildRecipeTable()
	buildOreNameTable()
	--refresh external designs?
end

--handle the building being built
function onBuiltEventHandler(event)
--see if we have a matching entity to run this
	if not design_lookup then return end
	if not design_lookup[event.created_entity.name] then return end
end

function onPlayerMinedEventHandler(event)
	--filter out non-mod entities
	if not design_lookup then return end
	if not design_lookup[event.entity.name] then return end
	local furnace, timer = removeFurnaceGroupByNumber(event.entity.unit_number)
	if (furnace) then forcePrint("Furnace group disabled",game.players[event.player_index].force) end
end

function onRobotMinedEventHandler(event)
	--filter out non-mod entities
	if not design_lookup then return end
	if not design_lookup[event.entity.name] then return end
	local furnace, timer = removeFurnaceGroupByNumber(event.entity.unit_number)
	if (furnace) then forcePrint("Furnace group disabled",event.entity.force) end
end

function onEntityDiedEventHandler(event)
	if not design_lookup then return end
	if not design_lookup[event.entity.name] then return end
	local furnace, timer = removeFurnaceGroupByNumber(event.entity.unit_number)
	if (furnace) then forcePrint("Furnace group disabled",event.entity.force) end
end

function onTickEventHandler(event)
	--0.0.11 furnace handler
	for index=game.tick%360+1,#global.furnaces,360 do
		if (global.timers[index]) then
			local result, err = pcall(processFurnaceNew, global.timers[index])
			if (err) then
				compound_furnace_debug = true
				--debugPrint("Debug Message Activated due to error")
				--debugPrint("Failed to run furnace handler for furnace ")
				--debugPrint("Timer: "..game.tick%360)
				--debugPrint("Furnace Index: "..global.timers[game.tick%360])
				--dsd(global.timers,"timers.txt")
				--dsd(global.furnaces[global.timers[game.tick%360]],"furnace.txt")
				debugPrint(err)
			end
		end

	end
end

-- called when the player selects a group of compound entities
function onPlayerSelectedArea(event)
--make sure the tool is for the mod
	if (event.item ~="cf-selector") then return end
--quick check to prevent timer overflow	
	if (#global.timers >1080) then
		game.players[event.player_index].print("The current limit for furnaces (1080) has been reached")
		return
	end
	
	local list = {}--a list to parse for types
	local entity_links = {}--entries of a lookup_table to link unit_number back to a furnace
--parse the entitites passed to it in event.entities
	for _,v in ipairs(event.entities) do
		local furnace, timer = getFurnaceGroupFromUnitNumber(v.unit_number)
		if (design_lookup[v.name] and furnace== nil) then 
--filter out entities that aren't in the mod, create a table for each entity type found
			if (not list[v.name]) then 
	 				list[v.name] = {} 
			end
			table.insert(list[v.name],{unit = v.unit_number,entity = v})
--add entity to entity links
			entity_links[v.unit_number] = {entity = v}
		end
		if (furnace) then
--entity from existing group found, quit early. (We technically can attempt to make a complete furnace group but using getFurnaceGroupFromUnitNumber is expensive with many furnaces)
			game.players[event.player_index].print("Selection overlaps with group "..timer)
			return 
		end
	end
	--debugPrint(serpent.dump(list))
	
	core = nil
	core_ref = nil
	furn_type = nil
	num = 0
-- find the design and see if there's only one core
	for i,v in ipairs(designs) do
		if (list[v.core]) then
			 if (core == nil) then 
				 core = v.core
			 core_ref = list[v.core][1].entity
			 furn_type = i
			 num = num + #list[v.core]
			 else 
				 game.players[event.player_index].print("More than one core type found!")
			 return 
		 end
		 end
	end
	
	--debugPrint(serpent.dump(list[core])) 
--continue if we found only 1 core
	if (num ~= 1) then 
		if (num == 0) then forcePrint("No Core Found",event.player_index) end
	if (num > 1) then forcePrint("More than one core found",event.player_index) end
		return 
	end	
	
	local design = designs[furn_type]
	
--start assembling the compound furnace structure
	compound = {}
	compound.furnace_design = furn_type
	compound.category = design.category
	
--get base entity counts
	if (list[design.batch]) then compound.batch = #list[design.batch] else compound.batch = 0 end
	if (list[design.speed]) then compound.speed = #list[design.speed] else compound.speed = 0 end
	compound.core = core_ref
	
--continue if batch and speed entities are within bounds
	if (compound.batch < 0 or compound.batch > design.batch_limit) then 
		forcePrint("Too Many Batch Processor Units, Max is "..design.batch_limit,event.player_index)
	--return
	end
	if (compound.speed < 0 or compound.speed >design.speed_limit) then 
		forcePrint("Too Many Speed Enhancement Units, Max is "..design.speed_limit,event.player_index)
	--return
	end	
	
--parse input chests
	compound.inputs = {}
	for _,v in ipairs (design.inputs) do
		if (list[v]) then
			for _,v in ipairs(list[v]) do
				table.insert(compound.inputs,v)
			end
		end

	end
--parse output chests
	compound.outputs = {}
	for _,v in ipairs (design.outputs) do
		if (list[v]) then	
			for _,v in ipairs(list[v]) do
				table.insert(compound.outputs,v)
			end
		end
	end
	
--continue if input and output entities are in limits
	if (#compound.inputs == 0) then 
		forcePrint("Furnace group requires at least 1 input",event.player_index)
	return
	end
	if (#compound.outputs == 0) then 
		forcePrint("Furnace group requires at least 1 output",event.player_index)
	return
	end
	if (#compound.inputs > design.input_limit) then 
		forcePrint("Too Many Inputs, Max is "..design.input_limit,event.player_index)
	return
	end
	if (#compound.outputs >design.output_limit) then 
		forcePrint("Too Many Outputs, Max is "..design.output_limit,event.player_index)
	return
	end
	
--set ownership
	compound.force = game.players[event.player_index].force
	forcePrint("Initializing compound furnace",compound.force)
--calculate the crafting rate and power draw
	local useful_speed_beacons = 8+2*(compound.batch - 1)
	local applied_speed = compound.speed
	if applied_speed > useful_speed_beacons then 
		applied_speed = useful_speed_beacons 
		forcePrint("Only "..applied_speed.." beacons counted for this build",compound.force)
		compound.speed = applied_speed
	end
	compound.craft_rate_base = design.craft_base + (compound.speed * 4.7604/compound.batch)
	compound.craft_rate_total = compound.craft_rate_base * compound.batch
	compound.power_per_sec = design.batch_watt * compound.batch * compound.craft_rate_base + compound.speed * design.speed_watt + compound.batch * 5
	compound.power_per_sec_idle = compound.speed * design.speed_watt + compound.batch * 5
	forcePrint("Power (peak): "..compound.power_per_sec.." Power(idle): "..compound.power_per_sec_idle)
--initialize products and timers
	--overflow is required because we can't predict what the contents of the chest will look like
	compound.overflow_name = "none"--the item name
	compound.overflow_count = 0 --the item count
	compound.overflow_max = 17*compound.batch*3
	compound.batch_name = nil--the batch item name
	compound.batch_count = 0--the batch item count
	compound.timer = getFreeTimerSlot()
	compound.links = entity_links--links between entity and unit ID
	compound.productivity = design.productivity
--push data to global

	local furnace = global.furnaceNext
	
	global.timers[compound.timer] = furnace
	global.furnaces[furnace] = compound
	global.furnaceNext = global.furnaceNext + 1
	
	forcePrint("Compound Furnace Created: " .. compound.craft_rate_total.. "x craft rate, 20% productivity bonus.",compound.force)
	forcePrint("This furnace makes: ".. math.floor(compound.craft_rate_total*6/3.5).." iron plates in 6 seconds",compound.force)
end

--attempt to move overflow items to the output chest, return
--returns new values for 
function OverflowToOutputChest(furnace)
	if (not furnace.overflow_count or furnace.overflow_count == 0 or not furnace.overflow_name) then return "none",0 end
--make the item stack
	local item = {name = furnace.overflow_name, count = math.floor(furnace.overflow_count/#furnace.outputs)}
	local transferred = 0
	local added = 0
	--transfer evenly
	for i,v in ipairs(furnace.outputs) do
		temp = v.entity.get_inventory(1).insert(item) 
		transferred = transferred + temp
	end
	--transfer unevenly
	item.count = furnace.overflow_count - transferred

	for i,v in ipairs(furnace.outputs) do
		
		if (item.count>0) then 
			added = v.entity.get_inventory(1).insert(item) 
			transferred = transferred + added
			item.count = item.count - added
		
		end
	end
	--update flow statistics
	furnace.force.item_production_statistics.on_flow(item.name,transferred)
	return item.name, item.count
end

--add batch to existing overflow, or make a new overflow
--returns true if the batch slot is empty at the end of the process
function BatchToOverflowChest(furnace)
	--existing case
	if (not furnace.batch_name or furnace.batch_count == 0 ) then return true end
	if (furnace.overflow_count == 0 or (furnace.overflow_name and furnace.overflow_name == furnace.batch_name)) then
		furnace.overflow_name = furnace.batch_name
		furnace.overflow_count = furnace.overflow_count + furnace.batch_count
		furnace.batch_count = 0
		return true
	end
	if (furnace.batch_count == 0) then 
		return true 
	else 
		return false 
	end
	
end

--get the maximum amount of input materials handled in a batch
function GetMaxInput(furnace,oreType)
	local recipe = global.recipes[oreType]
	
	return math.floor(furnace.craft_rate_total*6/recipe.energy)*recipe.ingredient.amount,recipe.ingredient.amount
end

--look in inputs and return the count and the name of the most plentiful ore
function FindBestOre(furnace)
	local oreNames = global.oreNames
	local counts = {}
	--counts table
	for a,b in ipairs(oreNames) do
		counts[a] = {oreNames[a],0}
	end
	--sum up the inventory
	for i,v in ipairs(furnace.inputs) do
		for j,k in ipairs(oreNames) do
			counts[j][2] = counts[j][2] + v.entity.get_inventory(defines.inventory.chest).get_item_count(k) 
		end
	end
	
	local bestName = nil
	local bestCount = 0
	--find best count
	for c,d in ipairs(oreNames) do
		if (bestCount < counts[c][2]) then
			bestName = counts[c][1]
			bestCount = counts[c][2]
		end
	end	
	return bestCount, bestName
end

--remove items from input inventories and return a count 
function PullFromInputs(furnace,ore,count)
	local transferred = 0
	local removed = 0
	local item = {name = ore, count = math.floor(count/#furnace.inputs)}
	--remove evenly
	for i,v in ipairs(furnace.inputs) do
		transferred = transferred + v.entity.get_inventory(1).remove(item)
	end
	item.count = count - transferred
	--remove unevenly
		for i,v in ipairs(furnace.inputs) do
			if (item.count > 0) then 
				removed =  v.entity.get_inventory(1).remove(item) 
				transferred = transferred + removed
				local temp = item.count - removed
				if (temp > 0) then
					item.count = temp
				else 
					return transferred
				end
			else
				return transferred;
			end
		end	
	return transferred
end


--make the batch and process statistics
function MakeBatch(furnace,count,oreName)
	local recipe = global.recipes[oreName]
	local rp = recipe.product
	local ri = recipe.ingredient
	local output_multi = rp.amount--default output multiplier
	if (not output_multi) then 
		--output multiplier to handle probabilistic quantities
		output_multi = (rp.amount_min+rp.amount_max)/2*rp.probability
	end
	local batch_count =  math.floor(count*(1+1/furnace.productivity)*output_multi/ri.amount)
	furnace.force.item_production_statistics.on_flow(oreName,count*-1)
	return batch_count, rp.name
end

function processFurnaceNew(index)
	local furnace = global.furnaces[index]
	local surface = furnace.core.surface
	local bestName = nil
	local bestCount = 0
	local amount_per_cycle = 0
	local max_count = 0
	local min_count = 0
	local count = 0
--move the batch to the overflow chest
	local batchEmpty = BatchToOverflowChest(furnace)
	furnace.overflow_name,furnace.overflow_count = OverflowToOutputChest(furnace)
	--set idle power drain
	if (batchEmpty and furnace.overflow_count < furnace.overflow_max) then 
--if theres space to make a new batch, see if we have enough for at least one run
		bestCount, bestName = FindBestOre(furnace)
		if (not bestCount or bestCount == 0 ) then 
			return
		end
--get next batch
		max_count,amount_per_cycle = GetMaxInput(furnace,bestName)
		min_count = amount_per_cycle*furnace.productivity
		count = math.floor(bestCount/min_count)*min_count
		if (count > max_count) then count = max_count end
		if (count < min_count) then 
			return 
		end;
--check if we have enough power		
		if (((furnace.power_per_sec_idle * 16.666666*360) + (furnace.power_per_sec * 16.666666 * count/ max_count)*360)>	furnace.core.energy) then
			--do we have a full enough power supply
			furnace.core.energy = furnace.core.energy - (furnace.power_per_sec_idle * 16.666666)*360
			return
		end		
--grab ore from inputs and run a batch
		local xferred = PullFromInputs(furnace,bestName,count)
		if (not (xferred == count)) 
		then 
			debugPrint("Furnace: "..index.." xferred error") 
			debugPrint("pfi "..xferred.." count "..count)
			return 
		end-- should be unreachable code
	
		furnace.batch_count, furnace.batch_name = MakeBatch(furnace,count,bestName)
--handle power draw and pollution		
		furnace.core.energy = furnace.core.energy - ((furnace.power_per_sec * 16.666666 * count/ max_count)*360)
		if (game.map_settings.pollution.enabled==true) then
			pollution_amt = ((furnace.power_per_sec - furnace.power_per_sec_idle) * count/max_count) * 6 / 300 *1.2
			--200Kj = 1 pollution, 1.2 = productivity
			surface.pollute(furnace.core.position,pollution_amt)
		end
	end
end

function buildRecipeTable()
	recipes = {}
	for i,v in pairs(game.recipe_prototypes) do
		--matches category
		if (v.category == "smelting") then
			data = {}
			--matches ingredient numbers
			if (#v.ingredients == 1 and #v.products == 1) then 
				data.ingredient = v.ingredients[1]
				data.product = v.products[1]
				data.energy = v.energy
		
				recipes[data.ingredient.name] = data
			end
		end
	end
	global.recipes = recipes
end

function buildOreNameTable()
	local index = 1
	oreNames = {}
	
	for i,v in pairs(game.recipe_prototypes) do
		--matches category
		if (v.category == "smelting") then
			--matches ingredient numbers
			if (#v.ingredients == 1 and #v.products == 1) then 
				oreNames[index] = v.ingredients[1].name
				index = index +1
			end
		end
	end
	global.oreNames = oreNames
end

--returns furnace index, timer index
function getFurnaceGroupFromUnitNumber(number)
	for i,v in pairs(global.timers) do
	furnace = global.furnaces[v]
		if (furnace.links[number]) then
			return v, furnace.timer 
		end
	end
	return nil,nil
end

function removeFurnaceGroupByNumber(number)
	local furnace,timer = getFurnaceGroupFromUnitNumber(number)
	
	if (furnace and timer) then 
		global.timers[timer] = nil
		global.furnaces[furnace] = nil
		return furnace, timer
	end
	
	return nil,nil
end

--this is a round robin timer. the on_tick routine touches only one furnace at a time (as a load balancing measure)
function getFreeTimerSlot()
	local timers = global.timers
		if (#timers == 1080) then return nil end --quick end if the timer database is full
	for i=1,1080,1 do
		if (not timers[i]) then return i end
	end
	return nil
end

function releaseTimerSlot(slot)
	global.timers[slot] = nil
end

function printFurnaceStats(id)
	local furnace = global.furnaces[id]
	game.write_file("test.txt",serpent.dump(furnace))
	debugPrint("Furnace no. "..id)
	debugPrint("Design: ".. designs[furnace.design].name)
	debugPrint("core: "..((furnace.core~=nil) and furnace.core.name or "none"))
	debugPrint("speed: "..furnace.speed_sum.."\\"..furnace.speed_limit)
	debugPrint("batch: "..furnace.batch_sum.."\\"..furnace.batch_limit)
	debugPrint("inputs: "..furnace.input_sum.."\\8")
	debugPrint("outputs: "..furnace.output_sum.."\\8")
end

function debugPrint(message)
	if not message then return end
	if compound_furnace_debug	then
		for i=1,#game.players,1 do
			game.players[i].print("Compound Furnaces: " .. message)
			--game.players[i].print(message)
						--game.write_file("cs_log.txt",game.tick.." Compound Furnaces: " .. message.."\n",true)
		end
	end
end

function dsd(message,file)
	if not message then return end
	if compound_furnace_debug then
		for i=1,#game.players,1 do
			--game.players[i].print(game.tick.."!																																Compound Furnaces: " .. message)
			game.players[i].print(serpent.dump(message))
						if (file) then game.write_file(file,game.tick.." Compound Furnaces: " .. serpent.dump(message).."\n") end
		end
	end
end

function forcePrint(message, force)
	if not force then debugPrint("No Force: "..message) return end
	if (type(force)=='number') then 
		game.players[force].force.print("Compound Furnaces: "..message)
	else
		force.print("Compound Furnaces: "..message)
	end
end

function on_init_handler()
	global.furnaces = {}--furnace dataset
	global.timers = {}
	global.furnaceNext = 1--the id of the next furnace to be created (inserted into global.furnaces)
	global.recipes = {}
	global.oreNames = {}
	
	buildRecipeTable()
	buildOreNameTable()
end

function on_load_handler()
	compound_furnace_debug = settings.global["compound_furnace_debug"].value
	update_debug_status()
end

function onModSettingEventHandler(event)
	if (event.setting == "compound_furnace_debug") then 
		update_debug_status()
		print_debug_status()
	end
end

function update_debug_status()
		if (settings.global["compound_furnace_debug"].value==true) then
			compound_furnace_debug = true
			commands.add_command("dump_furnace_numbers","dumps the furnace keys to a file in script-out",dumpFurnaceNumbers)
		else
			commands.remove_command("dump_furnace_numbers")
			compound_furnace_debug = false
		end
end

function print_debug_status()
		if (settings.global["compound_furnace_debug"].value==true) then
			debugPrint("Debug Messages/Commands On")
		else
			debugPrint("Debug Messages/Commands Off")
		end
end

--dumps the key's of valid furnace entries
function dumpFurnaceNumbers(name)
	output = "Furnaces "..#global.furnaces.." "
	local numbers = {}
	for i,v in pairs(global.furnaces) do
		table.insert(numbers,i)
	end
	game.write_file("Furnace_Numbers.lua",output..serpent.dump(numbers),false)
end

script.on_configuration_changed( 
function(data)
	local oldVersion
	local newVersion
	if data.mod_changes and data.mod_changes["compound-furnaces"] then 
		
		local index = 1
		if (data.mod_changes["compound-furnaces"].old_version) then
			oldVersion = {}
			for w in string.gmatch(data.mod_changes["compound-furnaces"].old_version,"%d+") do
				oldVersion[index] = tonumber(w)
				index = index + 1
			end
		end
		if (data.mod_changes["compound-furnaces"].new_version) then
			index = 1
			newVersion = {}
			for w in string.gmatch(data.mod_changes["compound-furnaces"].new_version,"%d+") do
				newVersion[index] = tonumber(w)
				index = index + 1
			end		
		end
	else
		return
	end

	--update beacon settings
	if  not(oldVersion ~=nil and 
	    oldVersion[1] and
		oldVersion[2] and 
		oldVersion[3] and
		oldVersion[1] >= 0 and 
		oldVersion[2] >= 0 and 
		oldVersion[3] >= 9 ) then
		game.print("Compound Splitters: updating furnaces to 0.0.9")
		for i,v in pairs(global.furnaces) do
			local useful_speed_beacons = 8+2*(v.batch - 1)
			local applied_speed = v.speed
			if applied_speed > useful_speed_beacons then 
				applied_speed = useful_speed_beacons 
				v.speed = applied_speed
			end		
		end
		game.print("Compound Splitters: "..#global.furnaces.." furnaces updated to 0.0.9")


	end
end
)


script.on_init( function() on_init_handler() end)
script.on_load( function() on_load_handler() end)
script.on_event(defines.events.on_built_entity, onBuiltEventHandler)
script.on_event(defines.events.on_tick, onTickEventHandler)
script.on_event(defines.events.on_player_mined_entity, onPlayerMinedEventHandler)
script.on_event(defines.events.on_entity_died, onEntityDiedEventHandler)
script.on_event(defines.events.on_robot_mined_entity,onRobotMinedEventHandler)
script.on_event(defines.events.on_player_selected_area, onPlayerSelectedArea)
script.on_event(defines.events.on_runtime_mod_setting_changed, onModSettingEventHandler)